
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: GetExpansionLevel, GetMaxLevelForExpansionLevel, LibStub, next, select, string

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local AddOnName, AddOnTable = ...

local AddOn = LibStub("AceAddon-3.0"):NewAddon(AddOnName, "AceEvent-3.0")
local Border = LibStub("SixBorder-1.0"):CreateFactory({})
local Constants = {}
local Database = {}

AddOnTable[1] = AddOn
AddOnTable[2] = Border
AddOnTable[3] = Constants
AddOnTable[4] = Database

_G[AddOnName] = AddOnTable

-------------------------------------------------------------------------------
-- AddOn
-------------------------------------------------------------------------------

AddOn:SetDefaultModuleLibraries("AceEvent-3.0", "AceHook-3.0", "AceTimer-3.0")
AddOn:SetDefaultModulePrototype({
	OnInitialize = function (self)
		self:RegisterMessage("SIXUI_CONFIG_CHANGED")
		self:LoadConfig()
	end,
	LoadConfig = function (self)
		self.config = AddOn:GetModuleConfig(self:GetName())
	end,
	SIXUI_CONFIG_CHANGED = function (self)
		self:LoadConfig()
	end
})
AddOn:SetDefaultModuleState(false)

function AddOn:OnInitialize()
	self:CreateDatabase()
	self:UpdateResources()
	self:UpdateBorder()

	-- Create a queue for callbacks that are waiting for combat to end.
	self.CombatQueue = LibStub("SixQueue-1.0"):CreateQueue()

	-- Create a queue for callbacks that are waiting for a frame to exist.
	self.FrameQueue = LibStub("SixQueue-1.0"):CreateQueue()
end

function AddOn:OnEnable()
	self:RegisterEvent("PLAYER_LEVEL_CHANGED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("UPDATE_EXPANSION_LEVEL")
	self:RegisterMessage("SIXUI_CONFIG_UPDATE")
	self:RegisterMessage("SIXUI_PROFILE_UPDATE")
	self:UpdateModules()
end

-- Create the SixUI database and callbacks.
function AddOn:CreateDatabase()
	Database.db = LibStub("AceDB-3.0"):New("SixDB", Constants.DATABASE_DEFAULTS, true)
	Database.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	Database.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	Database.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
end

-- Disable a module.
function AddOn:DisableModule(name)
	self:SetModuleState(name, false)
end

-- Enable a module.
function AddOn:EnableModule(name)
	self:SetModuleState(name, true)
end

-- Generate a SixBorder config from the database.
function AddOn:GetBorderConfig()
	return {
		accentBackdropInfo = {
			edgeFile = [[Interface\Buttons\WHITE8x8]],
			tile = true,
			tileSize = 8,
			tileEdge = true,
			edgeSize = Database.db.profile.border.accent.width,
		},
		accentAnchorOffset = Database.db.profile.border.accent.anchorOffset,
		accentBackdropBorderColor = {Constants.BORDER_ACCENT_COLOR:GetRGBA()},
		backgroundAnchorOffset = Database.db.profile.border.background.inset,
		backgroundBackdropColor = {Constants.BORDER_BACKGROUND_COLOR:GetRGBA()},
		strokeBackdropInfo = {
			edgeFile = [[Interface\Buttons\WHITE8x8]],
			tile = true,
			tileSize = 8,
			tileEdge = true,
			edgeSize = Database.db.profile.border.stroke.width,
		},
		strokeAnchorOffset = Database.db.profile.border.stroke.anchorOffset,
		strokeBackdropBorderColor = {Constants.BORDER_STROKE_COLOR:GetRGBA()},
	}
end

-- Get a module's config.
function AddOn:GetModuleConfig(name)
	return Database.db.profile.module and Database.db.profile.module[name] or {}
end

-- Check whether a module is configured to be disabled.
function AddOn:IsModuleDisabled(name)
	return not self:IsModuleEnabled(name)
end

-- Check whether a module is configured to be enabled.
function AddOn:IsModuleEnabled(name)
	return self:GetModuleConfig(name).enabled ~= false
end

-- Set the enabled state of a module.
function AddOn:SetModuleState(name, state)
	local module, config = self:GetModule(name), self:GetModuleConfig(name)

	if not module then
		AddOn.PrintF("The module '%s' wasn't recognised.", AddOn.Highlight(name))
	else
		config.enabled = state
		module:LoadConfig()
		module[state and "Enable" or "Disable"](module)
		self:GetModule("Miscellaneous Clutter"):Broadcast()
	end
end

-- Synchronise the database with SixUI's border color.
function AddOn:UpdateBorder()
	Border:SetConfig(self:GetBorderConfig())
end

-- Set the enabled state of all modules to reflect its config.
function AddOn:UpdateModules()
	for name in self:IterateModules() do
		self:SetModuleState(name, self:IsModuleEnabled(name))
	end
end

-- Synchronise the database with SixUI's resources.
function AddOn:UpdateResources()
	for key in next, Constants.CONFIG_KEY_TO_RESOURCE do
		local c, g, k = string.split(".", key)
		local v = Database.db.profile[c][g][k]

		if v then
			if (c == "border" and k == "color") or c == "color" then
				self:SIXUI_CONFIG_UPDATE(nil, key, v.r, v.g, v.b, v.a)
			elseif c == "font" then
				self:SIXUI_CONFIG_UPDATE(nil, key, v)
			end
		end
	end
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

function AddOn:OnProfileChanged()
	self:SendMessage("SIXUI_CONFIG_CHANGED")
	self:SendMessage("SIXUI_PROFILE_UPDATE")
end

function AddOn:PLAYER_LEVEL_CHANGED(event, oldLevel, newLevel, real)
	if real then
		Constants.PLAYER_LEVEL = newLevel
	end
end

function AddOn:PLAYER_REGEN_ENABLED()
	self.CombatQueue:Process()
end

function AddOn:SIXUI_CONFIG_UPDATE(event, key, ...)
	local c, g, k = string.split(".", key)

	if (c == "border" and k == "color") or c == "color" then
		if Constants.CONFIG_KEY_TO_RESOURCE[key] then
			Constants.CONFIG_KEY_TO_RESOURCE[key]:SetRGBA(...)
		end
	end

	if c == "border" then
		self:UpdateBorder()
	end

	if c == "font" and (g == "glyphFontSize" or g == "textFontSize") then
		local fontObject = Constants.CONFIG_KEY_TO_RESOURCE[key]

		if fontObject then
			local fontFile, _, fontFlags = fontObject:GetFont()

			fontObject:SetFont(fontFile, select(1, ...), fontFlags)
		end
	end
end

function AddOn:SIXUI_PROFILE_UPDATE()
	self:UpdateResources()
	self:UpdateBorder()
	self:UpdateModules()
end

function AddOn:UPDATE_EXPANSION_LEVEL()
	Constants.PLAYER_MAX_LEVEL = GetMaxLevelForExpansionLevel(GetExpansionLevel())
end
